@@ -1,6 +1,5 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
-/* $Id: items.c 257 2006-03-23 06:41:44Z bradfitz $ */
-
+/* $Id: items.c 450 2006-11-26 21:33:47Z sgrimm $ */
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -20,13 +19,14 @@
 
 #include "memcached.h"
 
-
-/* 
- * NOTE: we assume here for simplicity that slab ids are <=32. That's true in 
- * the powers-of-2 implementation, but if that changes this should be changed too
+/*
+ * We only reposition items in the LRU queue if they haven't been repositioned
+ * in this many seconds. That saves us from churning on frequently-accessed
+ * items.
  */
+#define ITEM_UPDATE_INTERVAL 60
 
-#define LARGEST_ID 32
+#define LARGEST_ID 255
 static item *heads[LARGEST_ID];
 static item *tails[LARGEST_ID];
 unsigned int sizes[LARGEST_ID];
@@ -41,14 +41,32 @@ void item_init(void) {
 }
 
 
-item *item_alloc(char *key, int flags, time_t exptime, int nbytes) {
-    int ntotal, len;
+/*
+ * Generates the variable-sized part of the header for an object.
+ *
+ * key     - The key 
+ * nkey    - The length of the key
+ * flags   - key flags
+ * nbytes  - Number of bytes to hold value and addition CRLF terminator
+ * suffix  - Buffer for the "VALUE" line suffix (flags, size).
+ * nsuffix - The length of the suffix is stored here.
+ *
+ * Returns the total size of the header.
+ */
+int item_make_header(char *key, uint8_t nkey, int flags, int nbytes,
+                     char *suffix, int *nsuffix) {
+    *nsuffix = sprintf(suffix, " %u %u\r\n", flags, nbytes - 2);
+    return sizeof(item) + nkey + *nsuffix + nbytes;
+}
+ 
+item *item_alloc(char *key, size_t nkey, int flags, rel_time_t exptime, int nbytes) {
+    int nsuffix, ntotal;
     item *it;
     unsigned int id;
+    char suffix[40];
 
-    len = strlen(key) + 1; if(len % 4) len += 4 - (len % 4);
-    ntotal = sizeof(item) + len + nbytes;
-    
+    ntotal = item_make_header(key, nkey + 1, flags, nbytes, suffix, &nsuffix);
+ 
     id = slabs_clsid(ntotal);
     if (id == 0)
         return 0;
@@ -93,11 +111,12 @@ item *item_alloc(char *key, int flags, time_t exptime, int nbytes) {
     it->next = it->prev = it->h_next = 0;
     it->refcount = 0;
     it->it_flags = 0;
-    it->nkey = len;
+    it->nkey = nkey;
     it->nbytes = nbytes;
     strcpy(ITEM_key(it), key);
     it->exptime = exptime;
-    it->flags = flags;
+    memcpy(ITEM_suffix(it), suffix, nsuffix);
+    it->nsuffix = nsuffix;
     return it;
 }
 
@@ -106,7 +125,7 @@ void item_free(item *it) {
     assert((it->it_flags & ITEM_LINKED) == 0);
     assert(it != heads[it->slabs_clsid]);
     assert(it != tails[it->slabs_clsid]);
-    assert(it->refcount == 0);    
+    assert(it->refcount == 0);
 
     /* so slab size changer can tell later if item is already free or not */
     it->slabs_clsid = 0;
@@ -114,9 +133,21 @@ void item_free(item *it) {
     slabs_free(it, ntotal);
 }
 
+/*
+ * Returns true if an item will fit in the cache (its size does not exceed
+ * the maximum for a cache entry.)
+ */
+int item_size_ok(char *key, size_t nkey, int flags, int nbytes) {
+    char prefix[40];
+    int nsuffix;
+
+    return slabs_clsid(item_make_header(key, nkey + 1, flags, nbytes,
+                                        prefix, &nsuffix)) != 0;
+}
+
 void item_link_q(item *it) { /* item is the new head */
     item **head, **tail;
-    assert(it->slabs_clsid <= LARGEST_ID);
+    /* always true, warns: assert(it->slabs_clsid <= LARGEST_ID); */
     assert((it->it_flags & ITEM_SLABBED) == 0);
 
     head = &heads[it->slabs_clsid];
@@ -134,10 +165,10 @@ void item_link_q(item *it) { /* item is the new head */
 
 void item_unlink_q(item *it) {
     item **head, **tail;
-    assert(it->slabs_clsid <= LARGEST_ID);
+    /* always true, warns: assert(it->slabs_clsid <= LARGEST_ID); */
     head = &heads[it->slabs_clsid];
     tail = &tails[it->slabs_clsid];
-    
+
     if (*head == it) {
         assert(it->prev == 0);
         *head = it->next;
@@ -159,8 +190,8 @@ int item_link(item *it) {
     assert((it->it_flags & (ITEM_LINKED|ITEM_SLABBED)) == 0);
     assert(it->nbytes < 1048576);
     it->it_flags |= ITEM_LINKED;
-    it->time = time(0);
-    assoc_insert(ITEM_key(it), it);
+    it->time = current_time;
+    assoc_insert(it);
 
     stats.curr_bytes += ITEM_ntotal(it);
     stats.curr_items += 1;
@@ -176,7 +207,7 @@ void item_unlink(item *it) {
         it->it_flags &= ~ITEM_LINKED;
         stats.curr_bytes -= ITEM_ntotal(it);
         stats.curr_items -= 1;
-        assoc_delete(ITEM_key(it));
+        assoc_delete(ITEM_key(it), it->nkey);
         item_unlink_q(it);
     }
     if (it->refcount == 0) item_free(it);
@@ -192,11 +223,13 @@ void item_remove(item *it) {
 }
 
 void item_update(item *it) {
-    assert((it->it_flags & ITEM_SLABBED) == 0);
+    if (it->time < current_time - ITEM_UPDATE_INTERVAL) {
+        assert((it->it_flags & ITEM_SLABBED) == 0);
 
-    item_unlink_q(it);
-    it->time = time(0);
-    item_link_q(it);
+        item_unlink_q(it);
+        it->time = current_time;
+        item_link_q(it);
+    }
 }
 
 int item_replace(item *it, item *new_it) {
@@ -207,7 +240,7 @@ int item_replace(item *it, item *new_it) {
 }
 
 char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int *bytes) {
-    
+
     int memlimit = 2*1024*1024;
     char *buffer;
     int bufcurr;
@@ -215,7 +248,7 @@ char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int
     int len;
     int shown = 0;
     char temp[512];
-    
+
     if (slabs_clsid > LARGEST_ID) return 0;
     it = heads[slabs_clsid];
 
@@ -224,7 +257,7 @@ char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int
     bufcurr = 0;
 
     while (it && (!limit || shown < limit)) {
-        len = sprintf(temp, "ITEM %s [%u b; %lu s]\r\n", ITEM_key(it), it->nbytes - 2, it->time);
+        len = sprintf(temp, "ITEM %s [%u b; %lu s]\r\n", ITEM_key(it), it->nbytes - 2, it->time + stats.started);
         if (bufcurr + len + 6 > memlimit)  /* 6 is END\r\n\0 */
             break;
         strcpy(buffer + bufcurr, temp);
@@ -235,7 +268,7 @@ char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int
 
     strcpy(buffer+bufcurr, "END\r\n");
     bufcurr+=5;
-    
+
     *bytes = bufcurr;
     return buffer;
 }
@@ -243,7 +276,7 @@ char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int
 void item_stats(char *buffer, int buflen) {
     int i;
     char *bufcurr = buffer;
-    time_t now = time(0);
+    rel_time_t now = current_time;
 
     if (buflen < 4096) {
         strcpy(buffer, "SERVER_ERROR out of memory");
@@ -252,7 +285,7 @@ void item_stats(char *buffer, int buflen) {
 
     for (i=0; i<LARGEST_ID; i++) {
         if (tails[i])
-            bufcurr += sprintf(bufcurr, "STAT items:%u:number %u\r\nSTAT items:%u:age %lu\r\n", 
+            bufcurr += sprintf(bufcurr, "STAT items:%u:number %u\r\nSTAT items:%u:age %u\r\n",
                                i, sizes[i], i, now - tails[i]->time);
     }
     strcpy(bufcurr, "END");
@@ -262,7 +295,7 @@ void item_stats(char *buffer, int buflen) {
 /* dumps out a list of objects of each size, with granularity of 32 bytes */
 char* item_stats_sizes(int *bytes) {
     int num_buckets = 32768;   /* max 1MB object, divided into 32 bytes size buckets */
-    unsigned int *histogram = (int*) malloc(num_buckets * sizeof(int));
+    unsigned int *histogram = (unsigned int*) malloc(num_buckets * sizeof(int));
     char *buf = (char*) malloc(1024*1024*2*sizeof(char));
     int i;
 
@@ -293,7 +326,32 @@ char* item_stats_sizes(int *bytes) {
         }
     }
     *bytes += sprintf(&buf[*bytes], "END\r\n");
-
     free(histogram);
     return buf;
 }
+
+/* expires items that are more recent than the oldest_live setting. */
+void item_flush_expired() {
+    int i;
+    item *iter, *next;
+    if (! settings.oldest_live)
+        return;
+    for (i = 0; i < LARGEST_ID; i++) {
+        /* The LRU is sorted in decreasing time order, and an item's timestamp
+         * is never newer than its last access time, so we only need to walk
+         * back until we hit an item older than the oldest_live time.
+         * The oldest_live checking will auto-expire the remaining items.
+         */
+        for (iter = heads[i]; iter != NULL; iter = next) {
+            if (iter->time >= settings.oldest_live) {
+                next = iter->next;
+                if ((iter->it_flags & ITEM_SLABBED) == 0) {
+                    item_unlink(iter);
+                }
+            } else {
+                /* We've hit the first old item. Continue to the next queue. */
+                break;
+            }
+        }
+    }
+}
